import logging
import multiprocessing
import time

from django.test import TestCase
from django.core.exceptions import ValidationError
from django_postgresql_dag.exceptions import NodeNotReachableException, GraphModelsCannotBeParsedException, IncorrectUsageException
from django_postgresql_dag.transformations import _filter_order, edges_from_nodes_queryset, nodes_from_edges_queryset, nx_from_queryset, model_to_dict

# from .dag_output import expected_dag_output
from .models import NetworkNode, NetworkEdge

logging.basicConfig(level=logging.DEBUG)


node_name_list = ["root", "a1", "a2", "a3", "b1", "b2", "b3", "b4", "c1", "c2"]


class DagTestCase(TestCase):
    def setUp(self):
        for node in node_name_list:
            NetworkNode.objects.create(name=node)

    def test_01_objects_were_created(self):
        log = logging.getLogger("test_01")
        log.debug("Creating objects")
        for node in node_name_list:
            self.assertEqual(NetworkNode.objects.get(name=f"{node}").name, f"{node}")
        log.debug("Done creating objects")

    def test_02_dag(self):
        log = logging.getLogger("test_02_dag")

        # Get nodes
        log.debug("Getting nodes")
        for node in node_name_list:
            globals()[f"{node}"] = NetworkNode.objects.get(name=node)
        log.debug("Done getting nodes")

        # Creates a DAG
        root.add_child(a1)
        b1.add_parent(a1)

        log.debug("descendants_tree")
        tree = root.descendants_tree()
        self.assertIn(a1, tree)
        self.assertEqual(len(tree), 1)
        self.assertIn(b1, tree[a1])
        self.assertEqual(tree[a1][b1], {})

        log.debug("descendants part 1")
        root_descendants = root.descendants()
        self.assertNotIn(root, root_descendants)
        self.assertTrue(all(elem in root_descendants for elem in [a1, b1]))

        root.add_child(a2)
        a3.add_parent(root)
        a3.add_child(b3)
        a3.add_child(b4)
        b3.add_child(c1)
        
        log.debug("descendants part 2")
        root_descendants = root.descendants()
        self.assertNotIn(root, root_descendants)
        self.assertTrue(all(elem in root_descendants for elem in [a1, a2, a3, b1, b3, b4, c1]))

        log.debug("ancestors part 1")
        c1_ancestors = c1.ancestors()
        self.assertNotIn(c1, c1_ancestors)
        self.assertNotIn(b4, c1_ancestors)
        self.assertTrue(all(elem in c1_ancestors for elem in [root, a3, b3]))
        
        a1.add_child(b2)
        a2.add_child(b2)
        b3.add_child(c2)
        b4.add_child(c1)
        
        log.debug("ancestors part 2")
        c1_ancestors = c1.ancestors()
        self.assertNotIn(c1, c1_ancestors)
        self.assertTrue(all(elem in c1_ancestors for elem in [root, a3, b3, b4]))
        
        # Try to add a node that is already an ancestor
        try:
            b3.add_parent(c1)
        except ValidationError as e:
            self.assertEqual(e.message, "The object is an ancestor.")

        # Try to add a node that is already an ancestor (alternate method)
        try:
            c1.add_child(b3)
        except ValidationError as e:
            self.assertEqual(e.message, "The object is an ancestor.")

        # Try to add a node as it's own child
        try:
            b3.add_child(b3)
        except ValidationError as e:
            self.assertEqual(e.message, "The object is an ancestor.")

        # Verify that the tree methods work
        log.debug("descendants_tree")
        tree_from_root = root.descendants_tree()
        self.assertIn(a1, tree_from_root)
        self.assertIn(a2, tree_from_root)
        self.assertIn(a3, tree_from_root)
        self.assertIn(b2, tree_from_root[a1])
        self.assertIn(b1, tree_from_root[a1])
        self.assertIn(b2, tree_from_root[a2])
        self.assertIn(b3, tree_from_root[a3])
        self.assertIn(b4, tree_from_root[a3])
        self.assertIn(c2, tree_from_root[a3][b3])
        self.assertIn(c1, tree_from_root[a3][b3])
        self.assertIn(c1, tree_from_root[a3][b4])

        self.assertEqual(len(tree_from_root), 3)
        self.assertEqual(len(tree_from_root[a3]), 2)
        self.assertEqual(len(tree_from_root[a3][b4]), 1)

        log.debug("ancestors_tree")
        tree_from_leaf = c1.ancestors_tree()
        self.assertIn(b3, tree_from_leaf)
        self.assertIn(a3, tree_from_leaf[b3])
        self.assertIn(b4, tree_from_leaf)
        self.assertIn(a3, tree_from_leaf[b4])
        self.assertIn(root, tree_from_leaf[b4][a3])

        self.assertEqual(len(tree_from_leaf), 2)
        self.assertEqual(len(tree_from_leaf[b3]), 1)
        self.assertEqual(len(tree_from_leaf[b4]), 1)
        self.assertEqual(len(tree_from_leaf[b4][a3]), 1)

        # Check other ancestor methods
        log.debug("ancestors_and_self")
        self.assertEqual(a1.ancestors_and_self()[0], root)
        log.debug("ancestors_and_self")
        self.assertEqual(a1.ancestors_and_self()[1], a1)
        log.debug("self_and_ancestors")
        self.assertEqual(a1.self_and_ancestors()[0], a1)
        log.debug("self_and_ancestors")
        self.assertEqual(a1.self_and_ancestors()[1], root)

        # Check other descendant methods
        log.debug("descendants_and_self")
        self.assertEqual(b4.descendants_and_self()[0], c1)
        log.debug("descendants_and_self")
        self.assertEqual(b4.descendants_and_self()[1], b4)
        log.debug("self_and_descendants")
        self.assertEqual(b4.self_and_descendants()[0], b4)
        log.debug("self_and_descendants")
        self.assertEqual(b4.self_and_descendants()[1], c1)

        # Check clan methods
        log.debug("clan_ids")
        self.assertTrue(all(elem in a1.clan() for elem in [root, a1, b1, b2]))
        log.debug("clan")
        self.assertEqual(a1.clan()[0], root)
        log.debug("clan")
        self.assertEqual(a1.clan()[3], b2)

        # Check distance between nodes
        log.debug("distance")
        self.assertEqual(root.distance(c1), 3)

        # Test additional fields for edge
        self.assertEqual(b3.children.through.objects.filter(child=c1)[0].name, "b3 c1")
        self.assertEqual(b3.descendants_edges().first(), NetworkEdge.objects.get(parent=b3, child=c1))
        self.assertEqual(a1.ancestors_edges().first(), NetworkEdge.objects.get(parent=root, child=a1))
        self.assertTrue(NetworkEdge.objects.get(parent=a1, child=b2) in a1.clan_edges())
        self.assertTrue(NetworkEdge.objects.get(parent=a1, child=b1) in a1.clan_edges())
        self.assertTrue(NetworkEdge.objects.get(parent=root, child=a1) in a1.clan_edges())

        # Test shortest_path
        log.debug("path x2")
        self.assertTrue(
            [p.name for p in root.path(c1)] == ["root", "a3", "b3", "c1"]
            or [p.name for p in c1.path(root, directional=False)]
            == ["root", "a3", "b4", "c1"]
        )

        log.debug("path")
        try:
            [p.name for p in c1.shortest_path(root)]
        except Exception as e:
            self.assertRaises(NodeNotReachableException)

        log.debug("shortest_path x2")
        self.assertTrue(
            [p.name for p in c1.path(root, directional=False)]
            == ["c1", "b3", "a3", "root"]
            or [p.name for p in c1.path(root, directional=False)]
            == ["c1", "b4", "a3", "root"]
        )

        log.debug("get_leaves")
        self.assertEqual([p.name for p in root.leaves()], ["b2", "c1", "c2", "b1"])
        log.debug("get_roots")
        self.assertEqual([p.name for p in c2.roots()], ["root"])

        self.assertTrue(root.is_root())
        self.assertTrue(c1.is_leaf())
        self.assertFalse(c1.is_root())
        self.assertFalse(root.is_leaf())
        self.assertFalse(a1.is_leaf())
        self.assertFalse(a1.is_root())

        # Remove a node and test island
        log.debug("descendants")
        self.assertTrue(c2 in b3.descendants())
        log.debug("ancestors")
        self.assertEqual([p.name for p in c2.ancestors()], ["root", "a3", "b3"])
        c2.remove_parent(b3)
        log.debug("descendants")
        self.assertFalse(c2 in b3.descendants())
        log.debug("ancestors")
        self.assertEqual([p.name for p in c2.ancestors()], [])
        self.assertTrue(c2.is_island())

        # Remove a node and test that it is still connected elsewhere
        log.debug("descendants")
        self.assertTrue(c1 in b3.descendants())
        log.debug("ancestors")
        self.assertEqual([p.name for p in c1.ancestors()], ["root", "a3", "b3", "b4"])
        b3.remove_child(c1)
        log.debug("descendants")
        self.assertFalse(c1 in b3.descendants())
        log.debug("ancestors")
        self.assertEqual([p.name for p in c1.ancestors()], ["root", "a3", "b4"])
        self.assertFalse(c1.is_island())
        
        # Test is we can properly export to a NetworkX graph
        log = logging.getLogger("test_02_networkx")
        nx_out = nx_from_queryset(c1.ancestors_and_self(), graph_attributes_dict={"test": "test"}, node_attribute_fields_list=["id", "name"], edge_attribute_fields_list=["id", "name"])
        log.debug("Check attributes")
        self.assertEqual(nx_out.graph, {"test": "test"})
        self.assertEqual(nx_out.nodes[11], {'id': 11, 'name': 'root'})
        self.assertEqual(nx_out.edges[11, 14], {'id': 4, 'name': 'root a3'})
        

        """
        Simulate a basic irrigation canal network
        """
        log = logging.getLogger("test_02_canal")

        node_name_list2 = [x for x in range(0, 201)]
        adjacency_list = [
            ["0", "1"],
            ["1", "2"],
            ["2", "3"],
            ["3", "4"],
            ["4", "5"],
            ["5", "6"],
            ["6", "7"],
            ["7", "8"],
            ["8", "9"],
            ["9", "10"],
            ["10", "11"],
            ["11", "12"],
            ["12", "13"],
            ["13", "14"],
            ["14", "15"],
            ["5", "16"],
            ["16", "17"],
            ["17", "18"],
            ["18", "19"],
            ["19", "20"],
            ["10", "21"],
            ["21", "22"],
            ["22", "23"],
            ["23", "24"],
            ["24", "25"],
            ["15", "26"],
            ["26", "27"],
            ["27", "28"],
            ["28", "29"],
            ["29", "30"],
            ["30", "31"],
            ["31", "32"],
            ["32", "33"],
            ["33", "34"],
            ["34", "35"],
            ["35", "36"],
            ["36", "37"],
            ["37", "38"],
            ["38", "39"],
            ["39", "40"],
            ["30", "41"],
            ["41", "42"],
            ["42", "43"],
            ["43", "44"],
            ["44", "45"],
            ["35", "46"],
            ["46", "47"],
            ["47", "48"],
            ["48", "49"],
            ["49", "50"],
            ["25", "51"],
            ["51", "52"],
            ["52", "53"],
            ["53", "54"],
            ["54", "55"],
            ["55", "56"],
            ["56", "57"],
            ["57", "58"],
            ["58", "59"],
            ["59", "60"],
            ["60", "61"],
            ["61", "62"],
            ["62", "63"],
            ["63", "64"],
            ["64", "65"],
            ["55", "66"],
            ["66", "67"],
            ["67", "68"],
            ["68", "69"],
            ["69", "70"],
            ["60", "71"],
            ["71", "72"],
            ["72", "73"],
            ["73", "74"],
            ["74", "75"],
            ["50", "76"],
            ["76", "77"],
            ["77", "78"],
            ["78", "79"],
            ["79", "80"],
            ["80", "81"],
            ["81", "82"],
            ["82", "83"],
            ["83", "84"],
            ["84", "85"],
            ["85", "86"],
            ["86", "87"],
            ["87", "88"],
            ["88", "89"],
            ["89", "90"],
            ["80", "91"],
            ["91", "92"],
            ["92", "93"],
            ["93", "94"],
            ["94", "95"],
            ["85", "96"],
            ["96", "97"],
            ["97", "98"],
            ["98", "99"],
            ["99", "100"],
            ["65", "101"],
            ["101", "102"],
            ["102", "103"],
            ["103", "104"],
            ["104", "105"],
            ["105", "106"],
            ["106", "107"],
            ["107", "108"],
            ["108", "109"],
            ["109", "110"],
            ["110", "111"],
            ["111", "112"],
            ["112", "113"],
            ["113", "114"],
            ["114", "115"],
            ["105", "116"],
            ["116", "117"],
            ["117", "118"],
            ["118", "119"],
            ["119", "120"],
            ["110", "121"],
            ["121", "122"],
            ["122", "123"],
            ["123", "124"],
            ["124", "125"],
            ["75", "126"],
            ["126", "127"],
            ["127", "128"],
            ["128", "129"],
            ["129", "130"],
            ["130", "131"],
            ["131", "132"],
            ["132", "133"],
            ["133", "134"],
            ["134", "135"],
            ["135", "136"],
            ["136", "137"],
            ["137", "138"],
            ["138", "139"],
            ["139", "140"],
            ["130", "141"],
            ["141", "142"],
            ["142", "143"],
            ["143", "144"],
            ["144", "145"],
            ["135", "146"],
            ["146", "147"],
            ["147", "148"],
            ["148", "149"],
            ["149", "150"],
            ["90", "151"],
            ["151", "152"],
            ["152", "153"],
            ["153", "154"],
            ["154", "155"],
            ["155", "156"],
            ["156", "157"],
            ["157", "158"],
            ["158", "159"],
            ["159", "160"],
            ["160", "161"],
            ["161", "162"],
            ["162", "163"],
            ["163", "164"],
            ["164", "165"],
            ["155", "166"],
            ["166", "167"],
            ["167", "168"],
            ["168", "169"],
            ["169", "170"],
            ["160", "171"],
            ["171", "172"],
            ["172", "173"],
            ["173", "174"],
            ["174", "175"],
            ["100", "176"],
            ["176", "177"],
            ["177", "178"],
            ["178", "179"],
            ["179", "180"],
            ["180", "181"],
            ["181", "182"],
            ["182", "183"],
            ["183", "184"],
            ["184", "185"],
            ["185", "186"],
            ["186", "187"],
            ["187", "188"],
            ["188", "189"],
            ["189", "190"],
            ["180", "191"],
            ["191", "192"],
            ["192", "193"],
            ["193", "194"],
            ["194", "195"],
            ["185", "196"],
            ["196", "197"],
            ["197", "198"],
            ["198", "199"],
            ["199", "200"],
        ]

        for n in range(1, 200):
            if n % 5 != 0:
                node_name_list2.append(f"SA{n}")
                node_name_list2.append(f"SB{n}")
                node_name_list2.append(f"SC{n}")

                adjacency_list.append([f"{n}", f"SA{n}"])
                adjacency_list.append([f"SA{n}", f"SB{n}"])
                adjacency_list.append([f"SA{n}", f"SC{n}"])

        # Create and assign nodes to variables
        log.debug("Start creating nodes")
        for node in node_name_list2:
            globals()[f"{node}"] = NetworkNode.objects.create(name=node)
        log.debug("Done creating nodes")


        log.debug("Connect nodes")
        for connection in adjacency_list:
            globals()[f"{connection[0]}"].add_child(globals()[f"{connection[1]}"])

        # Compute descendants of a root node
        canal_root = NetworkNode.objects.get(name="0")
        start_time = time.time()
        log.debug("Descendants: %s" % str(len(canal_root.descendants())))
        execution_time = time.time() - start_time
        log.debug("Execution time in seconds: %s" % str(execution_time))

        # Compute descendants of a leaf node
        canal_leaf = NetworkNode.objects.get(name="200")
        start_time = time.time()
        log.debug("Ancestors: %s" % str(len(canal_leaf.ancestors())))
        execution_time = time.time() - start_time
        log.debug("Execution time in seconds: %s" % str(execution_time))

        # Count number of paths from start to end of graph
        # NOTE: Does not work with current method of returning only a single path
        # start_time = time.time()
        # log.debug(
        #     "Paths through graph: : %s"
        #     % str(
        #         len(
        #             canal_root.path_ids_list(
        #                 canal_leaf, max_depth=n + 1, max_paths=500000000
        #             )
        #         )
        #     )
        # )
        # execution_time = time.time() - start_time
        # log.debug("Execution time in seconds: %s" % str(execution_time))

        # Find distance from root to leaf
        log.debug("Distance: %s" % str(canal_root.distance(canal_leaf, max_depth=100)))
        self.assertEqual(canal_root.distance(canal_leaf, max_depth=100), 60)

        log.debug("Node count: %s" % str(NetworkNode.objects.count()))
        log.debug("Edge count: %s" % str(NetworkEdge.objects.count()))

    def test_03_deep_dag(self):
        """
        Create a deep graph and check that graph operations run in a
        reasonable amount of time (linear in size of graph, not
        exponential).
        """

        def run_test():
            # Using the graph generation algorithm below, the number of potential
            # paths from node 0 doubles for each increase in n.
            # number_of_paths = 2^(n-1)  WRONG!!!
            # When n=22, there are on the order of 1 million paths through the graph
            # from node 0, so results for intermediate nodes need to be cached

            log = logging.getLogger("test_03")

            n = 22  # Keep it an even number


            log.debug("Start creating nodes")
            for i in range(2 * n):
                NetworkNode(pk=i, name=str(i)).save()
            log.debug("Done creating nodes")

            # Create edges
            log.debug("Connect nodes")
            for i in range(0, 2 * n - 2, 2):
                p1 = NetworkNode.objects.get(pk=i)
                p2 = NetworkNode.objects.get(pk=i + 1)
                p3 = NetworkNode.objects.get(pk=i + 2)
                p4 = NetworkNode.objects.get(pk=i + 3)

                p1.add_child(p3)
                p1.add_child(p4)
                p2.add_child(p3)
                p2.add_child(p4)

            # Compute descendants of a root node
            root_node = NetworkNode.objects.get(pk=0)
            start_time = time.time()
            log.debug("Descendants: %s" % str(len(root_node.ancestors())))
            execution_time = time.time() - start_time
            log.debug("Execution time in seconds: %s" % str(execution_time))

            # Compute ancestors of a leaf node
            leaf_node = NetworkNode.objects.get(pk=2 * n - 1)
            start_time = time.time()
            log.debug("Ancestors: %s" % str(len(leaf_node.ancestors())))
            execution_time = time.time() - start_time
            log.debug("Execution time in seconds: %s" % str(execution_time))

            first = NetworkNode.objects.get(name="0")
            last = NetworkNode.objects.get(pk=2 * n - 1)

            # # Count number of paths from start to end of graph
            # # NOTE: Does not work with current method of returning only a single path
            # start_time = time.time()
            # log.debug(
            #     "Paths through graph: %s"
            #     % str(
            #         len(first.path_ids_list(last, max_depth=n + 1, max_paths=500000000))
            #     )
            # )
            # execution_time = time.time() - start_time
            # log.debug("Execution time in seconds: %s" % str(execution_time))

            log.debug("Distance")
            self.assertEqual(first.distance(last, max_depth=n), n - 1)

            log.debug("Node count: %s" % str(NetworkNode.objects.count()))
            log.debug("Edge count: %s" % str(NetworkEdge.objects.count()))

            # Connect the first-created node to the last-created node
            NetworkNode.objects.get(pk=0).add_child(
                NetworkNode.objects.get(pk=2 * n - 1)
            )

            middle = NetworkNode.objects.get(pk=n - 1)
            log.debug("Distance")
            self.assertEqual(first.distance(middle, max_depth=n), n / 2 - 1)

        # Run the test, raising an error if the code times out
        p = multiprocessing.Process(target=run_test)
        p.start()
        p.join(20)  # Seconds allowed to live
        if p.is_alive():
            p.terminate()
            p.join()
            raise RuntimeError("Graph operations take too long!")




# DISALLOWED_ANCESTORS_NODES_CLAUSE_1 = """AND second.child_pk <> ALL(%(disallowed_ancestors_node_pks)s)"""  # Used for ancestors and upward path
# DISALLOWED_ANCESTORS_NODES_CLAUSE_2 = ("""AND {relationship_table}.child_pk <> ALL(%(disallowed_ancestors_node_pks)s)""")

# DISALLOWED_DESCENDANTS_NODES_CLAUSE_1 = """AND second.parent_pk <> ALL(%(disallowed_descendants_node_pks)s)"""  # Used for descendants and downward path
# DISALLOWED_DESCENDANTS_NODES_CLAUSE_2 = """AND {relationship_table}.parent_pk <> ALL(%(disallowed_descendants_node_pks)s)"""



CLUSTER_QUERY = """
WITH RECURSIVE traverse(path, last_parent, last_child) AS (
    SELECT ARRAY[id], id, id
    FROM {node_table} WHERE id = %(starting_node)s -- starting node
UNION ALL
    SELECT tv.path || edge.child_id || edge.parent_id, edge.profile1_id, edge.child_id
    FROM traverse tv
    JOIN {relationship_table} edge
    ON (edge.parent_id = tv.last_child AND NOT path @> array[edge.child_id]) 
       OR (edge.child_id = tv.last_parent AND NOT path @> array[edge.parent_id])
)
SELECT distinct unnest(path) FROM traverse;
"""




"""
A class to model hierarchies of objects following Directed Acyclic Graph structure.

The graph traversal queries use Postgresql's recursive CTEs to fetch an entire tree
of related node ids in a single query. These queries also topologically sort the ids
by generation.
"""

from django.apps import apps
from django.db import models, connection
from django.db.models import Case, When
from django.core.exceptions import ValidationError

from .exceptions import NodeNotReachableException
from .transformations import _ordered_filter
from .query_builders import (
    AncestorQuery,
    DescendantQuery,
    UpwardPathQuery,
    DownwardPathQuery,
    ConnectedGraphQuery
)


class NodeManager(models.Manager):
    def roots(self, node=None):
        """Returns a Queryset of all root Nodes, or optionally, the roots of a select node"""
        if node is not None:
            return node.roots()
        return self.filter(parents__isnull=True)

    def leaves(self, node=None):
        """Returns a Queryset of all leaf Nodes, or optionally, the leaves of a select node"""
        if node is not None:
            return node.leaves()
        return self.filter(children__isnull=True)


def node_factory(edge_model, children_null=True, base_model=models.Model):
    edge_model_table = edge_model._meta.db_table

    class Node(base_model):
        children = models.ManyToManyField(
            "self",
            blank=children_null,
            symmetrical=False,
            through=edge_model,
            related_name="parents",
        )

        objects = NodeManager()

        class Meta:
            abstract = True

        def get_foreign_key_field(self, fk_instance=None):
            """
            Provided a model instance, checks if the edge model has a ForeignKey field to the
            model class of that instance, and then returns the associated field name, else None.
            """
            if fk_instance is not None:
                for field in edge_model._meta.get_fields():
                    if field.related_model is fk_instance._meta.model:
                        # Return the first field that matches
                        return field.name
            return None

        def get_pk_name(self):
            """Sometimes we set a field other than 'pk' for the primary key.
            This method is used to get the correct primary key field name for the
            model so that raw queries return the correct information."""
            return self._meta.pk.name

        def ordered_queryset_from_pks(self, pks):
            """
            Generates a queryset, based on the current class and ordered by the provided pks
            """
            return _ordered_filter(self.__class__.objects, "pk", pks)

        def add_child(self, child, **kwargs):
            kwargs.update({"parent": self, "child": child})
            disable_check = kwargs.pop("disable_circular_check", False)
            cls = self.children.through(**kwargs)
            return cls.save(disable_circular_check=disable_check)

        def remove_child(self, child, delete_node=False):
            """Removes the edge connecting this node to child, and optionally deletes the child node as well"""
            if child in self.children.all():
                self.children.through.objects.get(parent=self, child=child).delete()
                if delete_node:
                    # Note: Per django docs:
                    # https://docs.djangoproject.com/en/dev/ref/models/instances/#deleting-objects
                    # This only deletes the object in the database; the Python instance will still
                    # exist and will still have data in its fields.
                    child.delete()

        def add_parent(self, parent, *args, **kwargs):
            return parent.add_child(self, **kwargs)

        def remove_parent(self, parent, delete_node=False):
            """Removes the edge connecting this node to parent, and optionally deletes the parent node as well"""
            if parent in self.parents.all():
                parent.children.through.objects.get(parent=parent, child=self).delete()
                if delete_node:
                    # Note: Per django docs:
                    # https://docs.djangoproject.com/en/dev/ref/models/instances/#deleting-objects
                    # This only deletes the object in the database; the Python instance will still
                    # exist and will still have data in its fields.
                    parent.delete()

        def ancestors_raw(self, **kwargs):
            return AncestorQuery(instance=self, **kwargs).raw_queryset()

        def ancestors(self, **kwargs):
            pks = [item.pk for item in self.ancestors_raw(**kwargs)]
            return self.ordered_queryset_from_pks(pks)

        def ancestors_count(self):
            return self.ancestors().count()

        def self_and_ancestors(self, **kwargs):
            pks = [self.pk] + [item.pk for item in self.ancestors_raw(**kwargs)][::-1]
            return self.ordered_queryset_from_pks(pks)

        def ancestors_and_self(self, **kwargs):
            pks = [item.pk for item in self.ancestors_raw(**kwargs)] + [self.pk]
            return self.ordered_queryset_from_pks(pks)

        def descendants_raw(self, **kwargs):
            return DescendantQuery(instance=self, **kwargs).raw_queryset()

        def descendants(self, **kwargs):
            pks = [item.pk for item in self.descendants_raw(**kwargs)]
            return self.ordered_queryset_from_pks(pks)

        def descendants_count(self):
            return self.descendants().count()

        def self_and_descendants(self, **kwargs):
            pks = [self.pk] + [item.pk for item in self.descendants_raw(**kwargs)]
            return self.ordered_queryset_from_pks(pks)

        def descendants_and_self(self, **kwargs):
            pks = [item.pk for item in self.descendants_raw(**kwargs)] + [self.pk]
            return self.ordered_queryset_from_pks(pks)

        def clan(self, **kwargs):
            """
            Returns a queryset with all ancestors, self, and all descendants
            """
            pks = (
                [item.pk for item in self.ancestors_raw(**kwargs)]
                + [self.pk]
                + [item.pk for item in self.descendants_raw(**kwargs)]
            )
            return self.ordered_queryset_from_pks(pks)

        def clan_count(self):
            return self.clan().count()

        def siblings(self):
            # Returns all nodes that share a parent with this node
            return self.siblings_with_self().exclude(pk=self.pk)

        def siblings_count(self):
            # Returns count of all nodes that share a parent with this node
            return self.siblings().count()

        def siblings_with_self(self):
            # Returns all nodes that share a parent with this node and self
            return self.__class__.objects.filter(
                parents__in=self.parents.all()
            ).distinct()

        def partners(self):
            # Returns all nodes that share a child with this node
            return self.partners_with_self().exclude(pk=self.pk)

        def partners_count(self):
            # Returns count of all nodes that share a child with this node
            return self.partners().count()

        def partners_with_self(self):
            # Returns all nodes that share a child with this node and self
            return self.__class__.objects.filter(
                children__in=self.children.all()
            ).distinct()

        def path_raw(self, ending_node, directional=True, max_depth=20, **kwargs):
            """
            Returns shortest path from self to ending node, optionally in either
            direction. The resulting RawQueryset is sorted from root-side, toward
            leaf-side, regardless of the relative position of starting and ending nodes.
            """

            if self == ending_node:
                return [[self.pk]]

            path = DownwardPathQuery(
                starting_node=self, ending_node=ending_node, **kwargs
            ).raw_queryset()

            if len(list(path)) == 0 and not directional:
                path = UpwardPathQuery(
                    starting_node=self, ending_node=ending_node, **kwargs
                ).raw_queryset()

            if len(list(path)) == 0:
                raise NodeNotReachableException

            return path

        def path(self, ending_node, **kwargs):
            pks = [item.pk for item in self.path_raw(ending_node, **kwargs)]
            return self.ordered_queryset_from_pks(pks)

        def distance(self, ending_node, **kwargs):
            """
            Returns the shortest hops count to the target node
            """
            return self.path(ending_node, **kwargs).count() - 1

        def is_root(self):
            """
            Check if has children and not ancestors
            """
            return bool(self.children.exists() and not self.parents.exists())

        def is_leaf(self):
            """
            Check if has ancestors and not children
            """
            return bool(self.parents.exists() and not self.children.exists())

        def is_island(self):
            """
            Check if has no ancestors nor children
            """
            return bool(not self.children.exists() and not self.parents.exists())

        def is_ancestor_of(self, ending_node, **kwargs):
            try:
                return len(self.path_raw(ending_node, **kwargs)) >= 1
            except NodeNotReachableException:
                return False

        def is_descendant_of(self, ending_node, **kwargs):
            return (
                not self.is_ancestor_of(ending_node, **kwargs)
                and len(self.path_raw(ending_node, directional=False, **kwargs)) >= 1
            )

        def is_sibling_of(self, ending_node):
            return ending_node in self.siblings()

        def is_partner_of(self, ending_node):
            return ending_node in self.partners()

        def node_depth(self):
            # Depth from furthest root
            # ToDo: Implement
            pass

        def connected_graph_raw(self, **kwargs):
            # Gets all nodes connected in any way to this node
            return ConnectedGraphQuery(instance=self, **kwargs).raw_queryset()

        def connected_graph(self, **kwargs):
            pks = [item.pk for item in self.connected_graph_raw(**kwargs)]
            return self.ordered_queryset_from_pks(pks)

        def descendants_tree(self):
            """
            Returns a tree-like structure with descendants
            # ToDo: Modify to use CTE
            """
            tree = {}
            for child in self.children.all():
                tree[child] = child.descendants_tree()
            return tree

        def ancestors_tree(self):
            """
            Returns a tree-like structure with ancestors
            # ToDo: Modify to use CTE
            """
            tree = {}
            for parent in self.parents.all():
                tree[parent] = parent.ancestors_tree()
            return tree

        def _roots(self, ancestors_tree):
            """
            Works on objects: no queries
            """
            if not ancestors_tree:
                return set([self])
            roots = set()
            for ancestor in ancestors_tree:
                roots.update(ancestor._roots(ancestors_tree[ancestor]))
            return roots

        def roots(self):
            """
            Returns roots nodes, if any
            # ToDo: Modify to use CTE
            """
            ancestors_tree = self.ancestors_tree()
            roots = set()
            for ancestor in ancestors_tree:
                roots.update(ancestor._roots(ancestors_tree[ancestor]))
            if len(roots) < 1:
                roots.add(self)
            return roots

        def _leaves(self, descendants_tree):
            """
            Works on objects: no queries
            """
            if not descendants_tree:
                return set([self])
            leaves = set()
            for descendant in descendants_tree:
                leaves.update(descendant._leaves(descendants_tree[descendant]))
            return leaves

        def leaves(self):
            """
            Returns leaves nodes, if any
            # ToDo: Modify to use CTE
            """
            descendants_tree = self.descendants_tree()
            leaves = set()
            for descendant in descendants_tree:
                leaves.update(descendant._leaves(descendants_tree[descendant]))
            if len(leaves) < 1:
                leaves.add(self)
            return leaves

        def descendants_edges(self):
            """
            Returns a queryset of descendants edges

            ToDo: Perform topological sort
            """
            return edge_model.objects.filter(
                parent__in=self.self_and_descendants(),
                child__in=self.self_and_descendants(),
            )

        def ancestors_edges(self):
            """
            Returns a queryset of ancestors edges

            ToDo: Perform topological sort
            """
            return edge_model.objects.filter(
                parent__in=self.self_and_ancestors(),
                child__in=self.self_and_ancestors(),
            )

        def clan_edges(self):
            """
            Returns a queryset of all edges associated with a given node
            """
            return self.ancestors_edges() | self.descendants_edges()

        @staticmethod
        def circular_checker(parent, child):
            if child in parent.self_and_ancestors():
                raise ValidationError("The object is an ancestor.")

    return Node



class EdgeManager(models.Manager):
    def from_nodes_queryset(self, nodes_queryset):
        """Provided a queryset of nodes, returns all edges where a parent and child
        node are within the queryset of nodes."""
        return _ordered_filter(self.model.objects, ["parent", "child"], nodes_queryset)

    def descendants(self, node, **kwargs):
        """
        Returns a queryset of all edges descended from the given node
        """
        return _ordered_filter(
            self.model.objects, "parent", node.self_and_descendants(**kwargs)
        )

    def ancestors(self, node, **kwargs):
        """
        Returns a queryset of all edges which are ancestors of the given node
        """
        return _ordered_filter(
            self.model.objects, "child", node.ancestors_and_self(**kwargs)
        )

    def clan(self, node, **kwargs):
        """
        Returns a queryset of all edges for ancestors, self, and descendants
        """
        return self.from_nodes_queryset(node.clan(**kwargs))

    def path(self, start_node, end_node, **kwargs):
        """
        Returns a queryset of all edges for the shortest path from start_node to end_node
        """
        return self.from_nodes_queryset(start_node.path(end_node, **kwargs))

    def validate_route(self, edges, **kwargs):
        """
        Given a list or set of edges, verify that they result in a contiguous route
        """
        # ToDo: Implement
        pass

    def sort(self, edges, **kwargs):
        """
        Given a list or set of edges, sort them from root-side to leaf-side
        """
        # ToDo: Implement
        pass


def edge_factory(
    node_model,
    concrete=True,
    base_model=models.Model,
):

    if isinstance(node_model, str):
        try:
            node_model_name = node_model.split(".")[1]
        except IndexError:

            node_model_name = node_model
    else:
        node_model_name = node_model._meta.model_name

    class Edge(base_model):
        parent = models.ForeignKey(
            node_model,
            related_name=f"{node_model_name}_child",
            on_delete=models.CASCADE,
        )
        child = models.ForeignKey(
            node_model,
            related_name=f"{node_model_name}_parent",
            on_delete=models.CASCADE,
        )

        objects = EdgeManager()

        class Meta:
            abstract = not concrete

        def save(self, *args, **kwargs):
            if not kwargs.pop("disable_circular_check", False):
                self.parent.__class__.circular_checker(self.parent, self.child)
            super().save(*args, **kwargs)

    return Edge



# Generated by Django 3.0.8 on 2020-10-19 02:55

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='NetworkEdge',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='NetworkNode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('children', models.ManyToManyField(blank=True, related_name='parents', through='tests.NetworkEdge', to='tests.NetworkNode')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='networkedge',
            name='child',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='NetworkNode_parent', to='tests.NetworkNode'),
        ),
        migrations.AddField(
            model_name='networkedge',
            name='parent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='NetworkNode_child', to='tests.NetworkNode'),
        ),
    ]



"""
Functions for transforming RawQuerySet or other outputs of
django-postgresql-dag to alternate formats.
"""

from django.core.exceptions import FieldDoesNotExist
from django.db.models import Case, When
from django.db.models.fields import DateTimeField, UUIDField
from django.db.models.fields.files import ImageField, FileField
from django.db.models.fields.related import ManyToManyField

from .exceptions import GraphModelsCannotBeParsedException, IncorrectUsageException

from itertools import chain
import networkx as nx


def _ordered_filter(queryset, field_names, values):
    """
    Filters the provided queryset for 'field_name__in values' for each given field_name in [field_names]
    orders results in the same order as provided values

        For instance
            _ordered_filter(self.__class__.objects, "pk", pks)
        returns a queryset of the current class, with instances where the 'pk' field matches an pk in pks

    """
    if not isinstance(field_names, list):
        field_names = [field_names]
    case = []
    for pos, value in enumerate(values):
        when_condition = {field_names[0]: value, "then": pos}
        case.append(When(**when_condition))
    order_by = Case(*case)
    filter_condition = {field_name + "__in": values for field_name in field_names}
    return queryset.filter(**filter_condition).order_by(order_by)


def get_instance_characteristics(instance):
    """
    Returns a tuple of the node & edge model classes and the instance_type
    for the provided instance
    """
    try:
        # Assume a queryset of nodes was provided
        _NodeModel = instance._meta.model
        _EdgeModel = instance._meta.model._meta.get_field("parents").through
        instance_type = "node"
    except FieldDoesNotExist:
        try:
            # Assume a queryset of edges was provided
            _EdgeModel = instance._meta.model
            _NodeModel = instance._meta.model._meta.get_field("parent").related_model
            instance_type = "edge"
        except FieldDoesNotExist:
            raise GraphModelsCannotBeParsedException
    return (_NodeModel, _EdgeModel, instance_type)


def get_queryset_characteristics(queryset):
    """
    Returns a tuple of the node & edge model classes and the queryset type
    for the provided queryset
    """
    try:
        # Assume a queryset of nodes was provided
        _NodeModel = queryset.model
        _EdgeModel = queryset.model._meta.get_field("parents").through
        queryset_type = "nodes_queryset"
    except FieldDoesNotExist:
        try:
            # Assume a queryset of edges was provided
            _EdgeModel = queryset.model
            _NodeModel = queryset.model._meta.get_field("parent").related_model
            queryset_type = "edges_queryset"
        except FieldDoesNotExist:
            raise GraphModelsCannotBeParsedException
    return (_NodeModel, _EdgeModel, queryset_type)


def model_to_dict(instance, fields=None, date_strf=None):
    """
    Returns a dictionary of {field_name: field_value} for a given model instance
    e.g.: model_to_dict(myqueryset.first(), fields=["id",])

    For DateTimeFields, a formatting string can be provided

    Adapted from: https://ziwon.github.io/post/using_custom_model_to_dict_in_django/
    """

    if not fields:
        raise IncorrectUsageException("fields list must be provided")

    opts = instance._meta
    data = {}
    __fields = list(map(lambda a: a.split("__")[0], fields or []))

    for f in chain(opts.concrete_fields, opts.private_fields, opts.many_to_many):
        is_editable = getattr(f, "editable", False)

        if fields and f.name not in __fields:
            continue

        if isinstance(f, DateTimeField):
            dt = f.value_from_object(instance)
            # Format based on format string provided, otherwise return a timestamp
            data[f.name] = dt.strftime(date_strf) if date_strf else dt.timestamp()

        elif isinstance(f, ImageField):
            image = f.value_from_object(instance)
            data[f.name] = image.url if image else None

        elif isinstance(f, FileField):
            file = f.value_from_object(instance)
            data[f.name] = file.url if file else None

        elif isinstance(f, ManyToManyField):
            if instance.pk is None:
                data[f.name] = []
            else:
                qs = f.value_from_object(instance)
                if qs._result_cache is not None:
                    data[f.name] = [item.pk for item in qs]
                else:
                    try:
                        m2m_field = list(
                            filter(lambda a: f.name in a and a.find("__") != -1, fields)
                        )[0]
                        key = m2m_field[len(f.name) + 2 :]
                        data[f.name] = list(qs.values_list(key, flat=True))
                    except IndexError:
                        data[f.name] = list(qs.values_list("pk", flat=True))

        if isinstance(f, UUIDField):
            uuid = f.value_from_object(instance)
            data[f.name] = str(uuid) if uuid else None

        # ToDo: Process other model fields

        elif is_editable:
            data[f.name] = f.value_from_object(instance)

    funcs = set(__fields) - set(list(data.keys()))
    for func in funcs:
        obj = getattr(instance, func)
        if inspect.ismethod(obj):
            data[func] = obj()
        else:
            data[func] = obj
    return data


def edges_from_nodes_queryset(nodes_queryset):
    """Given an Edge Model and a QuerySet or RawQuerySet of nodes,
    returns a queryset of the associated edges"""
    _NodeModel, _EdgeModel, queryset_type = get_queryset_characteristics(nodes_queryset)

    if queryset_type == "nodes_queryset":
        return _ordered_filter(_EdgeModel.objects, ["parent", "child"], nodes_queryset)
    raise IncorrectQuerysetTypeException


def nodes_from_edges_queryset(edges_queryset):
    """Given a Node Model and a QuerySet or RawQuerySet of edges,
    returns a queryset of the associated nodes"""
    _NodeModel, _EdgeModel, queryset_type = get_queryset_characteristics(edges_queryset)

    if queryset_type == "edges_queryset":

        nodes_list = (
            _ordered_filter(
                _NodeModel.objects,
                [
                    f"{_NodeModel.__name__}_child",
                ],
                edges_queryset,
            )
            | _ordered_filter(
                _NodeModel.objects,
                [
                    f"{_NodeModel.__name__}_parent",
                ],
                edges_queryset,
            )
        ).values_list("pk")

        return _NodeModel.objects.filter(pk__in=nodes_list)
    raise IncorrectQuerysetTypeException


def nx_from_queryset(
    queryset,
    graph_attributes_dict=None,
    node_attribute_fields_list=None,
    edge_attribute_fields_list=None,
    date_strf=None,
):
    """
    Provided a queryset of nodes or edges, returns a NetworkX graph

    Optionally, the following can be supplied to add attributes to components of the generated graph:
    graph_attributes_dict: A dictionary of attributes to add to the graph itself
    node_attribute_fields_list: a list of strings of field names to be added to nodes
    edge_attribute_fields_list: a list of strings of field names to be added to edges
    """
    _NodeModel, _EdgeModel, queryset_type = get_queryset_characteristics(queryset)

    if graph_attributes_dict is None:
        graph_attributes_dict = {}

    graph = nx.Graph(**graph_attributes_dict)

    if queryset_type == "nodes_queryset":
        nodes_queryset = queryset
        edges_queryset = edges_from_nodes_queryset(nodes_queryset)
    else:
        edges_queryset = queryset
        nodes_queryset = nodes_from_edges_queryset(edges_queryset)

    for node in nodes_queryset:
        if node_attribute_fields_list is not None:
            node_attribute_fields_dict = model_to_dict(
                node, fields=node_attribute_fields_list, date_strf=date_strf
            )
        else:
            node_attribute_fields_dict = {}

        graph.add_node(node.pk, **node_attribute_fields_dict)

    for edge in edges_queryset:
        if edge_attribute_fields_list is not None:
            edge_attribute_fields_dict = model_to_dict(
                edge, fields=edge_attribute_fields_list, date_strf=date_strf
            )
        else:
            edge_attribute_fields_dict = {}

        graph.add_edge(edge.parent.pk, edge.child.pk, **edge_attribute_fields_dict)

    return graph



# Generated by Django 3.0.8 on 2020-11-01 16:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='EdgeSet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='NodeSet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
            ],
        ),
        migrations.AddField(
            model_name='networkedge',
            name='edge_set',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='tests.EdgeSet'),
        ),
        migrations.AddField(
            model_name='networknode',
            name='edge_set',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='tests.EdgeSet'),
        ),
    ]



from abc import ABC, abstractmethod
from django.core.exceptions import ImproperlyConfigured
from .transformations import get_instance_characteristics, get_queryset_characteristics


class BaseQuery(ABC):
    """
    Base Query Class
    """

    def __init__(
        self,
        instance=None,
        starting_node=None,
        ending_node=None,
        max_depth=20,
        limiting_nodes_set_fk=None,
        limiting_edges_set_fk=None,
        disallowed_nodes_queryset=None,
        disallowed_edges_queryset=None,
        allowed_nodes_queryset=None,
        allowed_edges_queryset=None,
    ):
        self.instance = instance
        self.starting_node = starting_node
        self.ending_node = ending_node
        self.max_depth = max_depth
        self.limiting_nodes_set_fk = limiting_nodes_set_fk
        self.limiting_edges_set_fk = limiting_edges_set_fk
        self.disallowed_nodes_queryset = disallowed_nodes_queryset
        self.disallowed_edges_queryset = disallowed_edges_queryset
        self.allowed_nodes_queryset = allowed_nodes_queryset
        self.allowed_edges_queryset = allowed_edges_queryset

        if self.instance is not None:
            self.query_parameters = {
                "pk": self.instance.pk,
                "max_depth": self.max_depth,
            }
            (
                self.node_model,
                self.edge_model,
                self.instance_type,
            ) = get_instance_characteristics(self.instance)
        elif self.starting_node is not None and self.ending_node is not None:
            self.query_parameters = {
                "starting_node": self.starting_node.pk,
                "ending_node": self.ending_node.pk,
                "max_depth": self.max_depth,
            }
            (
                self.node_model,
                self.edge_model,
                self.instance_type,
            ) = get_instance_characteristics(self.starting_node)
        else:
            raise ImproperlyConfigured(
                "Either instance or both starting_node and ending_nod are required"
            )

        self.edge_model_table = self.edge_model._meta.db_table
        super().__init__()

    def limit_to_nodes_set_fk(self):
        """
        Limits the search to those nodes which are included in a ForeignKey's node set
            ToDo: Currently fails in the case that the starting node is not in the
              set of nodes related by the ForeignKey, but is adjacend to one that is
        """
        if not self.limiting_nodes_set_fk:
            return
        else:
            return self._limit_to_nodes_set_fk()

    @abstractmethod
    def _limit_to_nodes_set_fk(self):
        """Helper method. Override this method in subclasses"""
        return

    def limit_to_edges_set_fk(self):
        """
        Limits the search to those nodes which connect to edges defined in a ForeignKey's edge set
            ToDo: Currently fails in the case that the starting node is not in the
              set of nodes related by the ForeignKey, but is adjacend to one that is
        """
        if not self.limiting_edges_set_fk:
            return
        else:
            return self._limit_to_edges_set_fk()

    @abstractmethod
    def _limit_to_edges_set_fk(self):
        """Helper method. Override this method in subclasses"""
        return

    def disallow_nodes(self):
        """
        A queryset of Nodes that MUST NOT be included in the query
        """
        if not self.disallowed_nodes_queryset:
            return
        else:
            return self._disallow_nodes()

    @abstractmethod
    def _disallow_nodes(self):
        """Helper method. Override this method in subclasses"""
        return

    def disallow_edges(self):
        """
        A queryset of Edges that MUST NOT be included in the query
        """
        if not self.disallowed_edges_queryset:
            return
        else:
            return self._disallow_edges()

    @abstractmethod
    def _disallow_edges(self):
        """Helper method. Override this method in subclasses"""
        return

    def allow_nodes(self):
        """
        A queryset of Edges that MAY be included in the query
        """
        if not self.allowed_nodes_queryset:
            return
        else:
            return self._allow_nodes()

    @abstractmethod
    def _allow_nodes(self):
        """Helper method. Override this method in subclasses"""
        return

    def allow_edges(self):
        """
        A queryset of Edges that MAY be included in the query
        """
        if not self.allowed_edges_queryset:
            return
        else:
            return self._allow_edges()

    @abstractmethod
    def _allow_edges(self):
        """Helper method. Override this method in subclasses"""
        return

    @abstractmethod
    def raw_queryset(self):

        # Set the clases here, rather than in init so that we don't keep adding to the
        # clauses each time we check/utilize raw_queryset()
        self.where_clauses_part_1 = ""
        self.where_clauses_part_2 = ""

        self.limit_to_nodes_set_fk()
        self.limit_to_edges_set_fk()
        self.disallow_nodes()
        self.disallow_edges()
        self.allow_nodes()
        self.allow_edges()

        return

    def id_list(self):
        return [item.pk for item in self.raw_queryset()]

    def __str__(self):
        return str(self.raw_queryset())

    def __repr__(self):
        return str(self.raw_queryset())


class AncestorQuery(BaseQuery):
    """Ancestor Query Class"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.instance:
            raise ImproperlyConfigured("AncestorQuery requires an instance")
        return

    def _limit_to_nodes_set_fk(self):
        return

    def _limit_to_edges_set_fk(self):
        LIMITING_EDGES_SET_FK_CLAUSE_1 = (
            """AND second.{fk_field_name}_id = %(limiting_edges_set_fk_pk)s"""
        )
        LIMITING_EDGES_SET_FK_CLAUSE_2 = """AND {relationship_table}.{fk_field_name}_id = %(limiting_edges_set_fk_pk)s"""

        fk_field_name = get_foreign_key_field(fk_instance=self.limiting_edges_set_fk)
        if fk_field_name is not None:
            self.where_clauses_part_1 += "\n" + LIMITING_EDGES_SET_FK_CLAUSE_1.format(
                relationship_table=self.edge_model_table,
                # pk_name=self.instance.get_pk_name(),
                fk_field_name=fk_field_name,
            )
            self.where_clauses_part_2 += "\n" + LIMITING_EDGES_SET_FK_CLAUSE_2.format(
                relationship_table=self.edge_model_table,
                # pk_name=self.instance.get_pk_name(),
                fk_field_name=fk_field_name,
            )
            self.query_parameters[
                "limiting_edges_set_fk_pk"
            ] = self.limiting_edges_set_fk.pk

        return

    def _disallow_nodes(self):
        DISALLOWED_NODES_CLAUSE_1 = (
            """AND first.parent_id <> ALL(%(disallowed_node_pks)s)"""
        )
        DISALLOWED_NODES_CLAUSE_2 = (
            """AND {relationship_table}.parent_id <> ALL(%(disallowed_node_pks)s)"""
        )

        self.where_clauses_part_1 += "\n" + DISALLOWED_NODES_CLAUSE_1.format(
            relationship_table=self.edge_model_table,
            # pk_name=self.instance.get_pk_name(),
        )
        self.where_clauses_part_2 += "\n" + DISALLOWED_NODES_CLAUSE_2.format(
            relationship_table=self.edge_model_table,
            # pk_name=self.instance.get_pk_name(),
        )
        self.query_parameters["disallowed_node_pks"] = str(
            set(self.disallowed_nodes_queryset.values_list("pk", flat=True))
        )

        return

    def _disallow_edges(self):
        return

    def _allow_nodes(self):
        ALLOWED_NODES_CLAUSE_1 = """AND first.parent_id = ANY(%(allowed_node_pks)s)"""
        ALLOWED_NODES_CLAUSE_2 = (
            """AND {relationship_table}.parent_id = ANY(%(allowed_node_pks)s)"""
        )

        self.where_clauses_part_1 += "\n" + ALLOWED_NODES_CLAUSE_1.format(
            relationship_table=self.edge_model_table,
            # pk_name=self.instance.get_pk_name(),
        )
        self.where_clauses_part_2 += "\n" + ALLOWED_NODES_CLAUSE_2.format(
            relationship_table=self.edge_model_table,
            # pk_name=self.instance.get_pk_name(),
        )
        self.query_parameters["allowed_node_pks"] = str(
            set(self.allowed_nodes_queryset.values_list("pk", flat=True))
        )

        return

    def _allow_edges(self):
        return

    def raw_queryset(self):
        super().raw_queryset()

        QUERY = """
        WITH RECURSIVE traverse({pk_name}, depth) AS (
            SELECT first.parent_id, 1
                FROM {relationship_table} AS first
                LEFT OUTER JOIN {relationship_table} AS second
                ON first.parent_id = second.child_id
            WHERE first.child_id = %(pk)s
            -- LIMITING_FK_EDGES_CLAUSE_1
            -- DISALLOWED_ANCESTORS_NODES_CLAUSE_1
            -- ALLOWED_ANCESTORS_NODES_CLAUSE_1
            {where_clauses_part_1}
        UNION
            SELECT DISTINCT parent_id, traverse.depth + 1
                FROM traverse
                INNER JOIN {relationship_table}
                ON {relationship_table}.child_id = traverse.{pk_name}
            WHERE 1 = 1
            -- LIMITING_FK_EDGES_CLAUSE_2
            -- DISALLOWED_ANCESTORS_NODES_CLAUSE_2
            -- ALLOWED_ANCESTORS_NODES_CLAUSE_2
            {where_clauses_part_2}
        )
        SELECT {pk_name} FROM traverse
        WHERE depth <= %(max_depth)s
        GROUP BY {pk_name}
        ORDER BY MAX(depth) DESC, {pk_name} ASC
        """

        return self.node_model.objects.raw(
            QUERY.format(
                relationship_table=self.edge_model_table,
                pk_name=self.instance.get_pk_name(),
                where_clauses_part_1=self.where_clauses_part_1,
                where_clauses_part_2=self.where_clauses_part_2,
            ),
            self.query_parameters,
        )


class DescendantQuery(BaseQuery):
    """Descendant Query Class"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.instance:
            raise ImproperlyConfigured("DescendantQuery requires an instance")
        return

    def _limit_to_nodes_set_fk(self):
        return

    def _limit_to_edges_set_fk(self):
        LIMITING_EDGES_SET_FK_CLAUSE_1 = (
            """AND second.{fk_field_name}_id = %(limiting_edges_set_fk_pk)s"""
        )
        LIMITING_EDGES_SET_FK_CLAUSE_2 = """AND {relationship_table}.{fk_field_name}_id = %(limiting_edges_set_fk_pk)s"""

        fk_field_name = get_foreign_key_field(fk_instance=self.limiting_edges_set_fk)
        if fk_field_name is not None:
            self.where_clauses_part_1 += "\n" + LIMITING_EDGES_SET_FK_CLAUSE_1.format(
                relationship_table=self.edge_model_table,
                # pk_name=self.instance.get_pk_name(),
                fk_field_name=fk_field_name,
            )
            self.where_clauses_part_2 += "\n" + LIMITING_EDGES_SET_FK_CLAUSE_2.format(
                relationship_table=self.edge_model_table,
                # pk_name=self.instance.get_pk_name(),
                fk_field_name=fk_field_name,
            )
            self.query_parameters[
                "limiting_edges_set_fk_pk"
            ] = self.limiting_edges_set_fk.pk

        return

    def _disallow_nodes(self):
        DISALLOWED_NODES_CLAUSE_1 = (
            """AND first.child_id <> ALL(%(disallowed_node_pks)s)"""
        )
        DISALLOWED_NODES_CLAUSE_2 = (
            """AND {relationship_table}.child_id <> ALL(%(disallowed_node_pks)s)"""
        )

        self.where_clauses_part_1 += "\n" + DISALLOWED_NODES_CLAUSE_1.format(
            relationship_table=self.edge_model_table,
            # pk_name=self.instance.get_pk_name(),
        )
        self.where_clauses_part_2 += "\n" + DISALLOWED_NODES_CLAUSE_2.format(
            relationship_table=self.edge_model_table,
            # pk_name=self.instance.get_pk_name(),
        )
        self.query_parameters["disallowed_node_pks"] = str(
            set(self.disallowed_nodes_queryset.values_list("pk", flat=True))
        )

        return

    def _disallow_edges(self):
        return

    def _allow_nodes(self):
        ALLOWED_NODES_CLAUSE_1 = """AND first.child_id = ANY(%(allowed_node_pks)s)"""
        ALLOWED_NODES_CLAUSE_2 = (
            """AND {relationship_table}.child_id = ANY(%(allowed_node_pks)s)"""
        )

        self.where_clauses_part_1 += "\n" + ALLOWED_NODES_CLAUSE_1.format(
            relationship_table=self.edge_model_table,
            # pk_name=self.instance.get_pk_name(),
        )
        self.where_clauses_part_2 += "\n" + ALLOWED_NODES_CLAUSE_2.format(
            relationship_table=self.edge_model_table,
            # pk_name=self.instance.get_pk_name(),
        )
        self.query_parameters["allowed_node_pks"] = str(
            set(self.allowed_nodes_queryset.values_list("pk", flat=True))
        )

        return

    def _allow_edges(self):
        return

    def raw_queryset(self):
        super().raw_queryset()

        QUERY = """
        WITH RECURSIVE traverse({pk_name}, depth) AS (
            SELECT first.child_id, 1
                FROM {relationship_table} AS first
                LEFT OUTER JOIN {relationship_table} AS second
                ON first.child_id = second.parent_id
            WHERE first.parent_id = %(pk)s
            {where_clauses_part_1}
        UNION
            SELECT DISTINCT child_id, traverse.depth + 1
                FROM traverse
                INNER JOIN {relationship_table}
                ON {relationship_table}.parent_id = traverse.{pk_name}
            WHERE 1=1
            {where_clauses_part_2}
        )
        SELECT {pk_name} FROM traverse
        WHERE depth <= %(max_depth)s
        GROUP BY {pk_name}
        ORDER BY MAX(depth), {pk_name} ASC
        """

        return self.node_model.objects.raw(
            QUERY.format(
                relationship_table=self.edge_model_table,
                pk_name=self.instance.get_pk_name(),
                where_clauses_part_1=self.where_clauses_part_1,
                where_clauses_part_2=self.where_clauses_part_2,
            ),
            self.query_parameters,
        )


class ConnectedGraphQuery(BaseQuery):
    """Queries for the entire graph of nodes connected to the provided instance node"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.instance:
            raise ImproperlyConfigured("ConnectedGraphQuery requires an instance")
        return

    def _limit_to_nodes_set_fk(self):
        return

    def _limit_to_edges_set_fk(self):
        return

    def _disallow_nodes(self):
        return

    def _disallow_edges(self):
        return

    def _allow_nodes(self):
        return

    def _allow_edges(self):
        return

    def raw_queryset(self):
        super().raw_queryset()

        QUERY = """
        WITH RECURSIVE traverse AS
            (SELECT %(pk)s AS {pk_name}
            UNION SELECT
                CASE
                    WHEN edge.child_id = traverse.{pk_name} THEN edge.parent_id
                    ELSE edge.child_id
                END
            FROM traverse
            JOIN {relationship_table}. edge ON edge.parent_id = traverse.{pk_name}
            OR edge.child_id = traverse.{pk_name})
        SELECT *
        FROM traverse;
        """

        return self.node_model.objects.raw(
            QUERY.format(
                relationship_table=self.edge_model_table,
                pk_name=self.instance.get_pk_name(),
            ),
            self.query_parameters,
        )


class UpwardPathQuery(BaseQuery):
    """Upward Path Query Class"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.starting_node and not self.ending_node:
            raise ImproperlyConfigured(
                "UpwardPathQuery requires a starting_node and ending_node"
            )
        return

    def _limit_to_nodes_set_fk(self):
        return

    def _limit_to_edges_set_fk(self):
        LIMITING_EDGES_SET_FK_CLAUSE = (
            """AND first.{fk_field_name}_id = %(limiting_edges_set_fk_pk)s"""
        )

        fk_field_name = get_foreign_key_field(fk_instance=self.limiting_edges_set_fk)
        if fk_field_name is not None:
            self.where_clauses_part_2 += "\n" + LIMITING_EDGES_SET_FK_CLAUSE.format(
                relationship_table=self.edge_model_table,
                fk_field_name=fk_field_name,
            )
            self.query_parameters[
                "limiting_edges_set_fk_pk"
            ] = self.limiting_edges_set_fk.pk

        return

    def _disallow_nodes(self):
        DISALLOWED_NODES_CLAUSE = (
            """AND second.parent_id <> ALL('{disallowed_path_node_pks}')"""
        )

        self.where_clauses_part_2 += "\n" + DISALLOWED_NODES_CLAUSE
        self.query_parameters["disallowed_path_node_pks"] = str(
            set(self.disallowed_nodes_queryset.values_list("pk", flat=True))
        )

        return

    def _disallow_edges(self):
        return

    def _allow_nodes(self):
        ALLOWED_NODES_CLAUSE = (
            """AND second.parent_id = ALL('{allowed_path_node_pks}')"""
        )

        self.where_clauses_part_2 += "\n" + ALLOWED_NODES_CLAUSE
        self.query_parameters["allowed_path_node_pks"] = str(
            set(self.allowed_nodes_queryset.values_list("pk", flat=True))
        )

        return

    def _allow_edges(self):
        return

    def raw_queryset(self):
        super().raw_queryset()

        QUERY = """
        WITH RECURSIVE traverse(child_id, parent_id, depth, path) AS (
            SELECT
                first.child_id,
                first.parent_id,
                1 AS depth,
                ARRAY[first.child_id] AS path
                FROM {relationship_table} AS first
            WHERE child_id = %(starting_node)s
        UNION ALL
            SELECT
                first.child_id,
                first.parent_id,
                second.depth + 1 AS depth,
                path || first.child_id AS path
                FROM {relationship_table} AS first, traverse AS second
            WHERE first.child_id = second.parent_id
            AND (first.child_id <> ALL(second.path))
            -- PATH_LIMITING_FK_EDGES_CLAUSE
            -- DISALLOWED_UPWARD_PATH_NODES_CLAUSE
            -- ALLOWED_UPWARD_PATH_NODES_CLAUSE
            -- LIMITING_UPWARD_NODES_CLAUSE_1  -- CORRECT?
            {where_clauses_part_2}
        )
        SELECT 
            UNNEST(ARRAY[{pk_name}]) AS {pk_name}
        FROM 
            (
            SELECT path || ARRAY[%(ending_node)s], depth FROM traverse
                WHERE parent_id = %(ending_node)s
                AND depth <= %(max_depth)s
                LIMIT 1
        ) AS x({pk_name});
        """

        return self.node_model.objects.raw(
            QUERY.format(
                relationship_table=self.edge_model_table,
                pk_name=self.starting_node.get_pk_name(),
                where_clauses_part_2=self.where_clauses_part_2,
            ),
            self.query_parameters,
        )


class DownwardPathQuery(BaseQuery):
    """Downward Path Query Class"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.starting_node and not self.ending_node:
            raise ImproperlyConfigured(
                "DownwardPathQuery requires a starting_node and ending_node"
            )
        return

    def _limit_to_nodes_set_fk(self):
        return

    def _limit_to_edges_set_fk(self):
        LIMITING_EDGES_SET_FK_CLAUSE = (
            """AND first.{fk_field_name}_id = %(limiting_edges_set_fk_pk)s"""
        )

        fk_field_name = get_foreign_key_field(fk_instance=self.limiting_edges_set_fk)
        if fk_field_name is not None:
            self.where_clauses_part_2 += "\n" + LIMITING_EDGES_SET_FK_CLAUSE.format(
                relationship_table=self.edge_model_table,
                fk_field_name=fk_field_name,
            )
            self.query_parameters[
                "limiting_edges_set_fk_pk"
            ] = self.limiting_edges_set_fk.pk

        return

    def _disallow_nodes(self):
        DISALLOWED_NODES_CLAUSE = (
            """AND second.child_id <> ALL('{disallowed_path_node_pks}')"""
        )

        self.where_clauses_part_2 += "\n" + DISALLOWED_NODES_CLAUSE
        self.query_parameters["disallowed_path_node_pks"] = str(
            set(self.disallowed_nodes_queryset.values_list("pk", flat=True))
        )

        return

    def _disallow_edges(self):
        return

    def _allow_nodes(self):
        ALLOWED_NODES_CLAUSE = (
            """AND second.child_id = ALL('{allowed_path_node_pks}')"""
        )

        self.where_clauses_part_2 += "\n" + ALLOWED_NODES_CLAUSE
        self.query_parameters["allowed_path_node_pks"] = str(
            set(self.allowed_nodes_queryset.values_list("pk", flat=True))
        )

        return

    def _allow_edges(self):
        return

    def raw_queryset(self):
        super().raw_queryset()

        QUERY = """
        WITH RECURSIVE traverse(parent_id, child_id, depth, path) AS (
            SELECT
                first.parent_id,
                first.child_id,
                1 AS depth,
                ARRAY[first.parent_id] AS path
                FROM {relationship_table} AS first
            WHERE parent_id = %(starting_node)s
        UNION ALL
            SELECT
                first.parent_id,
                first.child_id,
                second.depth + 1 AS depth,
                path || first.parent_id AS path
                FROM {relationship_table} AS first, traverse AS second
            WHERE first.parent_id = second.child_id
            AND (first.parent_id <> ALL(second.path))
            -- PATH_LIMITING_FK_EDGES_CLAUSE
            -- DISALLOWED_DOWNWARD_PATH_NODES_CLAUSE
            -- ALLOWED_DOWNWARD_PATH_NODES_CLAUSE
            -- LIMITING_DOWNWARD_NODES_CLAUSE_1  -- CORRECT?
            {where_clauses_part_2}
        )      
        SELECT 
            UNNEST(ARRAY[{pk_name}]) AS {pk_name}
        FROM 
            (
            SELECT path || ARRAY[%(ending_node)s], depth FROM traverse
                WHERE child_id = %(ending_node)s
                AND depth <= %(max_depth)s
                LIMIT 1
        ) AS x({pk_name});
        """

        return self.node_model.objects.raw(
            QUERY.format(
                relationship_table=self.edge_model_table,
                pk_name=self.starting_node.get_pk_name(),
                where_clauses_part_2=self.where_clauses_part_2,
            ),
            self.query_parameters,
        )



#!/usr/bin/env python

import os
from setuptools import setup

version = '0.1.0'

classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Operating System :: OS Independent",
    "Topic :: Software Development :: Libraries",
    "Topic :: Database",
    "Topic :: Utilities",
    "Environment :: Web Environment",
    "Framework :: Django",
    "Framework :: Django :: 2.0",
    "Framework :: Django :: 2.1",
    "Framework :: Django :: 2.2",
    "Framework :: Django :: 3.0",
    "Framework :: Django :: 3.1",
]

root_dir = os.path.dirname(__file__)
if not root_dir:
    root_dir = '.'
long_desc = open(root_dir + '/README.md').read()

setup(
    name='django-postgresql-dag',
    version=version,
    url='https://github.com/OmenApps/django-postgresql-dag',
    author='Jack Linke, et al.',
    author_email='jacklinke@gmail.com',
    license='Apache Software License',
    packages=['django_postgresql_dag'],
    package_dir={'django_postgresql_dag': 'django_postgresql_dag'},
    description='Directed Acyclic Graph implementation for Django & Postgresql',
    classifiers=classifiers,
    long_description_content_type="text/markdown",
    long_description=long_desc,
)



"""Provides exceptions relevant to graphs"""


class NodeNotReachableException(Exception):
    """
    Exception for node distance and path
    """

    pass


class GraphModelsCannotBeParsedException(Exception):
    """
    Exception for node distance and path
    """

    pass


class IncorrectUsageException(Exception):
    """
    Exception for node distance and path
    """

    pass




from django.db import models
from django_postgresql_dag.models import node_factory, edge_factory


class EdgeSet(models.Model):
    """A model designed as a container for a set of edges"""
    name = models.CharField(max_length=100)


class NodeSet(models.Model):
    """A model designed as a container for a set of nodes"""
    name = models.CharField(max_length=100)


class NetworkEdge(edge_factory("NetworkNode", concrete=False)):
    name = models.CharField(max_length=100)

    edge_set = models.ForeignKey(EdgeSet, null=True, blank=True, on_delete=models.CASCADE)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        self.name = f"{self.parent.name} {self.child.name}"
        super().save(*args, **kwargs)

    class Meta:
        app_label = 'tests'


class NetworkNode(node_factory(NetworkEdge)):
    name = models.CharField(max_length=100)

    edge_set = models.ForeignKey(EdgeSet, null=True, blank=True, on_delete=models.CASCADE)

    def __str__(self):
        return self.name

    class Meta:
        app_label = 'tests'